#include "omf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char* OMF_record_type_name[] = {
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    "THEADR", NULL, "LHEADR", NULL, NULL, NULL, NULL, NULL,
    "COMENT", NULL, "MODEND", "MODEND", "EXTDEF", NULL, NULL, NULL,
    "PUBDEF", "PUBDEF", NULL, NULL, "LINNUM", "LINNUM", "LNAMES", NULL,
    "SEGDEF", "SEGDEF", "GRPDEF", NULL, "FIXUPP", "FIXUPP", NULL, NULL,
    "LEDATA", "LEDATA", "LIDATA", "LIDATA", NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    "COMDEF", NULL, "BAKPAT", "BAKPAT", "LEXTDEF", NULL, "LPUBDEF", "LPUBDEF",
    "LCOMDEF", NULL, NULL, NULL, "CEXTDEF", NULL, NULL, NULL,
    NULL, NULL, "COMDAT", "COMDAT", "LINSYM", "ALIAS", NULL, NULL,
    "NBKPAT", "NBKPAT", "LLNAMES", NULL, "VERNUM", NULL, "VENDEXT", NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    "LIBHED", "LIBEND", NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
};

int parse(OMF* omf, FILE* fp) {
    OMF_Record r;
    memset(&r, 0, sizeof(OMF_Record));

    r.type = fgetc(fp);
    if (r.type == EOF) return 0;

    if (fread(&r.size, 2, 1, fp) != 1) return 1;

    r.off = ftell(fp);

    r.cont = (char*)malloc(sizeof(char)*r.size);
    if ((int)fread(r.cont, 1, r.size, fp) != r.size) return 1;

    /* delete check sum and null delimit */
    r.size--;
    r.sum = r.cont[r.size];
    r.cont[r.size] = '\0';

    /* skip null */
    if (r.type == 0x8a) {
        int c;
//        printf("skip 0x8a\n");
        while ((c = fgetc(fp)) == 0);
        ungetc(c, fp);
    }

    /* set max size */
    if (r.type == 0xf0) {
        omf->dict_off = *(int*)r.cont;
        omf->dict_size = *(short*)(r.cont+4);
    }

    omf->record_size++;
    omf->records = (OMF_Record*)realloc(omf->records,
                                        sizeof(OMF_Record)*omf->record_size);
    omf->records[omf->record_size-1] = r;

    return 0;
}

OMF* OMF_load(const char* filename) {
    OMF* omf;
    FILE* fp;

    fp = fopen(filename, "rb");
    if (!fp) {
        printf("%s: file not found\n", filename);
        return NULL;
    }

    omf = (OMF*)malloc(sizeof(OMF));
    memset(omf, 0, sizeof(OMF));

    while (!feof(fp) && (omf->dict_off == 0 || ftell(fp) < omf->dict_off)) {
        if (parse(omf, fp)) {
            printf("%s: collapsed OMF file?\n", filename);
            free(omf);
            return NULL;
        }

        /* LIBEND */
        if (omf->records[omf->record_size-1].type == 0xf1) break;
    }

    if (omf->dict_size) {
        omf->dictionary = (char*)malloc(omf->dict_size*512);
        fread(omf->dictionary, 1, omf->dict_size*512, fp);
    }

    fclose(fp);

    return omf;
}

void OMF_free(OMF* omf) {
    int i;
    for (i = 0; i < omf->record_size; i++) {
        free(omf->records[i].cont);
    }
    free(omf->records);
    free(omf);
}

void OMF_save(OMF* omf, const char* filename) {
    FILE* fp;
    int i;

    fp = fopen(filename, "wb");
    for (i = 0; i < omf->record_size; i++) {
        OMF_Record r = omf->records[i];
        int s = r.off - ftell(fp) - 3;

        if (s > 0) {
            char buf[32];
            memset(buf, 0, s);
            fwrite(buf, 1, s, fp);
        }

        fputc(r.type, fp);

        /* LIBEND */
        if (i == omf->record_size-1 && r.type == 0xf1) {
            int off = ftell(fp);
            char buf[512];
            s = ((off+0x202)&~0x1ff) - off - 2;
            fwrite(&s, 2, 1, fp);
            memset(buf, 0, s);
            fwrite(buf, 1, s, fp);
            break;
        }

        s = r.size+1;
        fwrite(&s, 2, 1, fp);
        fwrite(r.cont, 1, r.size, fp);
        fputc(r.sum, fp);
    }

    if (omf->dict_size) {
//        int off = ftell(fp);
        fwrite(omf->dictionary, 1, omf->dict_size*512, fp);
//        fseek(fp, 3, SEEK_SET);
//        fwrite(&off, 4, 1, fp);
    }
    fclose(fp);
}
